#include "sun.hpp"
#include "scene.hpp"
#include "camera.hpp"
#include "shader.hpp"
#include "sceneObject.hpp"
#include "cubeMapObject.hpp"
#include "debug.hpp"
#include "lightCamera.hpp"
#include "sunVolumetric.hpp"
#include "mouseSemantics.hpp"

#include <GL/glew.h>
#include <GLFW/glfw3.h>

#include <string>
#include <iostream>
#include <stdexcept>
#include <sstream>

#define APP_NAME "Selektron 29.01.2015"
#define DEBUG 1 //TODO: discuss this

static void mouseCallback(GLFWwindow* window, double posx, double posy) {
	MouseSemantics *mouseSemantics = reinterpret_cast<MouseSemantics *> (glfwGetWindowUserPointer(window));
	//mouseSemantics->mousePosCallback(window,posx,posy);
}

int main(int argc, char** argv)
{
    int width = 1920;
    int height = 1080;

    GLFWwindow* window = NULL;
    std::unique_ptr<Scene> scene;
    std::unique_ptr<Debug> debug;
	std::unique_ptr<MouseSemantics> mouseSemantics;

    try {
        {//window creation
            int refresh = 60;
            bool fullscreen = true;
            if(argc == 2) {
                if(argv[1][0] == 'h') {
                    std::cout << APP_NAME << std::endl << "Arg list: widht, height, refresh, (0-windowed, 1-fullscreen)" << std::endl;
                    return 0;
                }
                /*width = 1280;
                height = 720;*/
                fullscreen = false;
            }

            if(argc == 5) {
                width = std::strtol(argv[1], 0, 10);
                height = std::strtol(argv[2], 0, 10);
                refresh = std::strtol(argv[3], 0, 10);
                fullscreen = std::strtol(argv[4], 0, 10) != 0;
            }

            if(!glfwInit())
                throw std::runtime_error("Error on glfwInit");

            GLFWmonitor* monitor = NULL;
            if(fullscreen) {
                monitor = glfwGetPrimaryMonitor();
                auto screen = glfwGetVideoMode(monitor);
                if(refresh == 0)
                    refresh = screen->refreshRate;
                glfwWindowHint(GLFW_REFRESH_RATE, refresh);
            }

            if(!(window = glfwCreateWindow(width, height, APP_NAME, monitor, NULL)))
                throw std::runtime_error("Error on glfwCreateWindow");
			glfwMakeContextCurrent(window);
            if(glewInit() != GLEW_OK)
                throw std::runtime_error("Error on glewInit");
            if(!GLEW_VERSION_4_3)
                throw std::runtime_error("OpenGL 4.3 is not supported");

			double xpos;
			double ypos;
			glfwGetCursorPos(window, &xpos, &ypos);
			mouseSemantics.reset(new MouseSemantics(xpos, ypos));
			glfwSetWindowUserPointer(window, mouseSemantics.get());
			glfwSetCursorPosCallback(window, mouseCallback);
            
			//TODO: discuss this
            #if DEBUG
            Debug::enableWindowsDebug();
            #endif
        }

        scene.reset(new Scene(window));

        std::cout << "Compile shaders " << std::endl;
        {//compile shaders
            std::unique_ptr<Shader> shader;
            std::vector<std::pair<GLenum, std::string> > shaderPaths;
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/shadowMap.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/shadowMap.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("shadowMap", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/particleCPU.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/particleCPU.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("particleCPU", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/particleTessGPU.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_CONTROL_SHADER, "../shader/particleTessGPU.tc.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_EVALUATION_SHADER, "../shader/particleTessGPU.te.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/particleTessGPU.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/phongColorOmni.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("particleTessGPU", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/particleTessGPU.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_CONTROL_SHADER, "../shader/particleTessGPU.tc.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_EVALUATION_SHADER, "../shader/particleTessGPUOmni.te.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/OmniFirst.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/OmniFirst.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("particleTessGPUOmni", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_COMPUTE_SHADER, "../shader/particleSystemGPU.cs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("particleSystemGPU", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/generic.vert"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/generic.frag"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("generic", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/phongTexture.frag"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("phongTexture", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/CMTexture.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/CMTexture.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/CMTexture.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("CMTexture", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/CMColor.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/CMColor.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/CMColor.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("CMColor", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/CMReflect.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("CMReflect", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/phongColor.frag"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("phongColor", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/OmniFirst.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/OmniFirst.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/OmniFirst.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("omniShadow", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/phongColorOmni.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/phongColorOmni.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("omniColor", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/phongTextureOmni.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/phongTextureOmni.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("omniTexture", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/SVAmbientColor.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("SVambientColor", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/SVAmbientTexture.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("SVambientTexture", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_GEOMETRY_SHADER, "../shader/SV.gs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/SV.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("SV", shader);
            shaderPaths.clear();
            //TODO: use for advanced volumetric lightning
            /*shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/OmniFirst.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("VolumeLightShadowPass", shader);
            shaderPaths.clear();*/
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/volumetricFirstPass.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("volumetricFirstPass", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/SV.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/volumetricFirstPassSun.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("volumetricFirstPassSun", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/volumetricPostProcess.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/volumetricPostProcess.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("volumetricPostProcess", shader);
            shaderPaths.clear();
            shaderPaths.push_back(std::make_pair(GL_VERTEX_SHADER, "../shader/tessellation.vs.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_CONTROL_SHADER, "../shader/tessellation.tc.glsl"));
            shaderPaths.push_back(std::make_pair(GL_TESS_EVALUATION_SHADER, "../shader/tessellation.te.glsl"));
            shaderPaths.push_back(std::make_pair(GL_FRAGMENT_SHADER, "../shader/tessellation.fs.glsl"));
            shader.reset(new Shader(shaderPaths));
            scene->addShader("tessellation", shader);
            shaderPaths.clear();

        }
        std::cout << "Shader compilation finished" << std::endl;

        scene->init("../demo.dae");
        debug.reset(new Debug(scene.get(), APP_NAME));
    } catch(std::exception& ex) {
        std::cout << "Error on init: " << ex.what() << std::endl;
        glfwTerminate();
        return -1;
    } catch(...) {
        std::cout << "Error on init: Unknown exception received" << std::endl;
        glfwTerminate();
        return -1;
    }
           
    //init effects
    scene->getSun()->reset();

    glEnable(GL_DEPTH_TEST);
    //glEnable(GL_TEXTURE_3D);

    // Enable blending
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    glViewport(0, 0, width, height);
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    bool cubeRender = true;
    bool omniShadow = true;
    bool firstTimeCube = true;
    //scene->setOmniShader();

    bool paused = false;
    bool running = true;
    double deltaT = 0.0;
    double lastTime = 0.0;
    while (!glfwWindowShouldClose(window) && glfwGetKey(window, GLFW_KEY_ESCAPE) != GLFW_PRESS)
    {//restart
        lastTime = 0;
        running = true;
        firstTimeCube = true;
        glfwSetTime(0.5);
        
        for(auto it = scene->begin(); it != scene->end(); ++it) {
            std::shared_ptr<SceneObject> object = Scene::get(it);

            object->setAnimationTime(0);
        }
    //main loop
    while (running && !glfwWindowShouldClose(window)) {
        double time = glfwGetTime();
        deltaT = time - lastTime;
        lastTime = time;

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        running = paused;
        debug->update(time, deltaT);
        scene->getCamera()->update(deltaT); //debug

        //animate all objects, if not paused
        for(auto it = scene->begin(); !paused && it != scene->end(); ++it) {
            std::shared_ptr<SceneObject> object = Scene::get(it);

            //must be bitwise OR,
            //when false function would not be called
            running = running | object->animate(time);
        }

        //std::cout << "Current time: " << time << std::endl;
        
        //arrange effects for omni directional shadows
        glEnable(GL_DEPTH_TEST);
        //glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        //glEnable(GL_CULL_FACE);
        //glCullFace(GL_BACK);
        scene->getSun()->drawOmni();
        //glCullFace(GL_BACK);
        
        //draw all objects
        for(auto it = scene->begin(); it != scene->end(); ++it) {
            std::shared_ptr<SceneObject> object = Scene::get(it);
            if(!object->doNotRender() && !object->getIsVolSun())
                object->draw();

        }

        //handle pausing
        paused = paused | (glfwGetKey(window, GLFW_KEY_SPACE) == GLFW_PRESS);
        paused = paused & (glfwGetKey(window, GLFW_KEY_ENTER) != GLFW_PRESS);

        //fast-forward is disabled
#if 0
        if(glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS || glfwGetKey(window, GLFW_KEY_A) == GLFW_PRESS) {
            static double lastSkip = time;
            if(time - lastSkip < 0)
                lastSkip = time; //handle restart
            if(time - lastSkip > 0.1) {
                if(glfwGetKey(window, GLFW_KEY_D) == GLFW_PRESS)
                    time = time + 100*deltaT;
                else
                    time = time - 200*deltaT;

                glfwSetTime(time);
                for(auto it = scene->begin(); it != scene->end(); ++it) {
                    std::shared_ptr<SceneObject> object = Scene::get(it);

                    object->setAnimationTime(time);
                }
                lastSkip = time;
            }
        }
#endif
        auto error = glGetError();
        if(error != GL_NO_ERROR)
            std::cerr << "OpenGL Error: " << glewGetErrorString(error) << std::endl;

        glfwSwapBuffers(window);
        glfwPollEvents();
        running = running && glfwGetKey(window, GLFW_KEY_ESCAPE) != GLFW_PRESS;//can be logical AND, if false we exit anyway
    } //main loop
    } //restart

    glfwDestroyWindow(window);
    glfwTerminate();
    return 0;
}

