#include "scene.hpp"
#include "camera.hpp"

#include <GL/glew.h>
#include <GLFW/glfw3.h>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtx/rotate_vector.hpp>
#include <iostream>

Camera::Camera(const std::string& name, Scene* scene, const glm::vec3& eye, float vAngle, float hAngle, float fov, float near, float far)
    : SceneObject(name, scene)
    , fov(fov)
    , farClip(far)
    , nearClip(near)
    , eye(eye)
{
    int width;
    int height;
    glfwGetWindowSize(scene->getWindow(), &width, &height);
    projectionMatrix = glm::perspective(fov, width / (float)height, nearClip, farClip);

    cursorX = width / 2.0;
    cursorY = height / 2.0;

    glm::vec3 direction(cos(vAngle) * sin(hAngle),
                        cos(vAngle) * cos(hAngle),
                        sin(vAngle));
    glm::vec3 right(sin(hAngle - 3.14f/2.0f),
                    cos(hAngle - 3.14f/2.0f),
                    0);
    glm::vec3 up = glm::cross(right, direction);

    viewMatrix = glm::lookAt(eye, eye + direction, up);
    modelMatrix = glm::inverse(viewMatrix);
}

Camera::Camera(const std::string& name, Scene* scene, glm::vec3& pos, glm::vec3& up, glm::vec3& eye, float fov, float near, float far)
    : SceneObject(name, scene)
    , fov(glm::degrees(fov))
    , farClip(far)
    , nearClip(near)
    , up(up)
    , eye(eye)
    , pos(pos)
{
    int width;
    int height;
    glfwGetWindowSize(scene->getWindow(), &width, &height);
    projectionMatrix = glm::perspective(glm::degrees(fov), width / (float)height, near, far);

	//this->eye = glm::vec3(getGlobalModelMatrix() * glm::vec4(eye,1));

    cursorX = width / 2.0;
    cursorY = height / 2.0;
}

void Camera::setViewMatrix(const glm::mat4& modelMatrix)
{
    this->modelMatrix = modelMatrix;
    glm::mat4 M = getGlobalModelMatrix();
    viewMatrix = glm::inverse(M);
}

Camera::~Camera()
{

}

void Camera::draw() const
{

}

void Camera::update(double deltaT)
{
	int width;
    int height;
    double xpos;
    double ypos;
    GLFWwindow* window = scene->getWindow();
    glfwGetCursorPos(window, &xpos, &ypos);
    glfwGetWindowSize(window, &width, &height);

	glm::vec4 c_pos = getGlobalModelMatrix() * glm::vec4(0, 0, 0, 1);
	//std::cout << "X Cam: " << c_pos.x << " Y Cam: " << c_pos.y << " Z Cam: " << c_pos.z << std::endl;

    if(glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_LEFT) == GLFW_PRESS)
    {//rotate
        //glfwSetCursorPos(window, cursorX, cursorY);
        float verticalAngle = (cursorY - ypos);
		float horizontalAngle = (cursorX - xpos);
		cursorX = xpos;
		cursorY = ypos;

		glm::vec4 c_pos = getGlobalModelMatrix() * glm::vec4(0,0,0,1);
		//rotate
		glm::vec3 lookat = eye;
		glm::vec3 view = glm::normalize(lookat - glm::vec3(c_pos));
		glm::vec3 upvec = glm::vec3(0, 1, 0);

		view = glm::rotate(view, (float)deltaT * 0.50f * horizontalAngle, upvec);

		glm::vec3 rotateAround = glm::cross(view, upvec);
		view = glm::rotate(view, (float)deltaT * 0.50f * verticalAngle, rotateAround);

		//up = glm::cross(rotateAround,eye);
		std::cout << "pos x: " << c_pos.x << " pos y: " << c_pos.y << "pos z: " << c_pos.z << std::endl;
//TODO position update gets messed up

		this->viewMatrix = glm::lookAt(glm::vec3(c_pos.x, c_pos.y, c_pos.z), glm::vec3(c_pos.x, c_pos.y, c_pos.z) + view, upvec);
		this->modelMatrix = glm::inverse(viewMatrix);

#if 0
		//This works but only for 0 0 0 
		this->viewMatrix = glm::lookAt(glm::vec3(0, 0, 0), glm::vec3(0, 0, 0) + lookat, upvec);
		this->modelMatrix = glm::inverse(viewMatrix);
#endif
#if 0
        viewMatrix = glm::inverse(modelMatrix);
        viewMatrix = glm::rotate(viewMatrix, (float)verticalAngle, glm::vec3(1, 0, 0));
        viewMatrix = glm::rotate(viewMatrix, (float)horizontalAngle, glm::vec3(0, 0, 1));
        setViewMatrix(glm::inverse(viewMatrix));
#endif
	} else if(glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_RIGHT) == GLFW_PRESS)
    {//move z
        glfwSetCursorPos(window, cursorX, cursorY);
        double delta = (cursorY - ypos);

        modelMatrix = glm::translate(modelMatrix, glm::vec3(0, 0, delta));
        setViewMatrix(modelMatrix);
    } else if(glfwGetKey(window, GLFW_KEY_UP) == GLFW_PRESS)
    {//move up
        modelMatrix = glm::translate(modelMatrix, glm::vec3(0, 10 * deltaT, 0));
        setViewMatrix(modelMatrix);
    } else if(glfwGetKey(window, GLFW_KEY_DOWN) == GLFW_PRESS)
    {//move down
        modelMatrix = glm::translate(modelMatrix, glm::vec3(0, -10 * deltaT, 0));
        setViewMatrix(modelMatrix);
    } else if(glfwGetKey(window, GLFW_KEY_LEFT) == GLFW_PRESS)
    {//move left
        modelMatrix = glm::translate(modelMatrix, glm::vec3(-10 * deltaT, 0, 0));
        setViewMatrix(modelMatrix);
    } else if(glfwGetKey(window, GLFW_KEY_RIGHT) == GLFW_PRESS)
    {//move right
        modelMatrix = glm::translate(modelMatrix, glm::vec3(10 * deltaT, 0, 0));
        setViewMatrix(modelMatrix);
    } else if(glfwGetKey(window, GLFW_KEY_1) == GLFW_PRESS)
    {//zoom out
        fov += float(10 * deltaT);
        projectionMatrix = glm::perspective(fov, width / (float)height, nearClip, farClip);
    } else if(glfwGetKey(window, GLFW_KEY_2) == GLFW_PRESS)
    {//zoom in
        fov -= float(10 * deltaT);
        projectionMatrix = glm::perspective(fov, width / (float)height, nearClip, farClip);
    } else {//no camera action
        cursorX = xpos;
        cursorY = ypos;
    }
}

bool Camera::animate(double time) {
#if 0
	bool ret = SceneObject::animate(time);
    setViewMatrix(modelMatrix);
    return ret;
#endif
	return true;
}
