#include "mouseSemantics.hpp"
#include <iostream>
#include <functional>
#include <math.h>

/*
int deltapos_x = floor((xpos - mCursorLastPos->x) * mCursorFactor);
int deltapos_y = floor((ypos - mCursorLastPos->y) * mCursorFactor);
mMousePosition->x = mCursorLastPos->x + deltapos_x;
mMousePosition->y = mCursorLastPos->y + deltapos_y;

*/

MouseSemantics::MouseSemantics(float _M, float _m, float _C, GLFWwindow* _window)
{
	latestPos = glm::vec2(-1, -1);
	M = _M;
	m = _m;
	C = _C;
	distance = 40;
	inWindow = false;
	firstTimeInWindow = false,
	window = _window;
}

void MouseSemantics::updateDistanceState(unsigned int _distance) {
	distance = (int) _distance;
}

void MouseSemantics::cursorEnterCallBack(GLFWwindow *window, int entered) {
	if (entered) {
		inWindow = true;
		firstTimeInWindow = true;
	}
	else {
		inWindow = false;
		firstTimeInWindow = false;
	}
}

void MouseSemantics::update(double deltaT) {

	double xpos;
	double ypos;
	glfwGetCursorPos(window, &xpos, &ypos);

	//disable movement scaling when moving camera
	if (glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_RIGHT) == GLFW_PRESS) {
		return;
	}

	if (glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_LEFT) == GLFW_PRESS) {
		return;
	}

	if (!inWindow) {
		return;
	}

	if (firstTimeInWindow) {
		firstTimeInWindow = false;
		latestPos.x = xpos;
		latestPos.y = ypos;
		return;
	}

	glm::vec2 direction = glm::vec2((float)xpos, (float)ypos) - latestPos;
	float dlength = glm::length(direction);
	std::cout << "Length: " << dlength << std::endl;
	std::cout << "Old Position x " << xpos << " Old position y " << ypos << std::endl;
	std::cout << "Direction x " << direction.x << " Direction y " << direction.y << std::endl;
	//std::cout << "distance: " << distance << std::endl;

	if (dlength > glm::epsilon<float>()) {
		//direction = direction / dlength;
		//scaling
		/*
		while (dlength > glm::epsilon<float>()) {
			float toTravel = glm::min(dlength, 1.0f);
			scaleAlongDir(window, direction, toTravel);
			dlength = dlength - toTravel;
			//std::cout << "Dlength " << dlength << std::endl;
		}
		*/
		//scaleAlongDir(window, direction, dlength);
		
		direction = direction;
		latestPos = latestPos + direction* 1.0f;

		glfwSetCursorPos(window, latestPos.x, latestPos.y);
		
		std::cout << "New Position x: " << latestPos.x << " New position y: " << latestPos.y << std::endl;
	}
}

void MouseSemantics::scaleAlongDir(GLFWwindow *window, glm::vec2 direction, float toTravel) {
#if 0
	if (C < distance) {
		direction = direction * M;
	}
	else if (0 < distance && distance < C) {
		direction = direction * (distance / C) * (M - m) + m;
		//std::cout << "speed " << ((distance / C) * (M - m) + m) << std::endl;
	}
	else {
		direction = direction * m;
	}
#endif
	latestPos = latestPos + direction * 0.1f;
	glfwSetCursorPos(window, latestPos.x, latestPos.y);
}

void MouseSemantics::mousePosCallback(GLFWwindow *window, double xpos, double ypos) {
	//x - horizontal pos
	//y - vertical pos

	//mouse cursor is not in window
	if (!inWindow) {
		return;
	}

	//disable movement scaling when moving camera
	if (glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_RIGHT) == GLFW_PRESS) {
		return;
	}
	if (glfwGetMouseButton(window, GLFW_MOUSE_BUTTON_LEFT) == GLFW_PRESS) {
		return;
	}

	//first calls where nothing happens
	if ((latestPos.x == -1 && latestPos.y == -1) || distance == -1) {
		latestPos.x = (float)xpos;
		latestPos.y = (float)ypos;
		return;
	}

	glm::vec2 direction = glm::vec2((float)xpos, (float)ypos) - latestPos;
	float dlength = glm::length(direction);
	std::cout << "Length: " << dlength << std::endl;
	std::cout << "distance: " << distance << std::endl;

	if (dlength > glm::epsilon<float>()) {
		//direction = direction / dlength;
		//scaling
#if 0
		while (dlength > glm::epsilon<float>()) {
			float toTravel = glm::min(dlength, 1.0f);
			scaleAlongDir(window, direction, toTravel);
			dlength = dlength - toTravel;
			//std::cout << "Dlength " << dlength << std::endl;
		}
#endif	
		//scaleAlongDir(window, direction, dlength);
		direction = direction * 5.4f;
		latestPos = latestPos + direction;

		//glfwSetCursorPos(window, latestPos.x, latestPos.y);
		std::cout << "New Position x: " << latestPos.x << "New position y: " << latestPos.y << std::endl;
	}
}
